# ifndef LCLMISC_H
# define LCLMISC_H

typedef	unsigned int bits;
typedef long unsigned Handle;
extern bool firstWord (const char *p_s, const char *p_w);
extern void fputline (FILE *p_out, char *p_s) /*@modifies p_out@*/;

# endif
