/*
** Copyright (C) University of Virginia, Massachusetts Institue of Technology 1994-2003.
** See ../LICENSE for license information.
**
*/

/*
** LSLInitRuleCode defined in lslinit.h
*/

typedef LSLInitRuleCode LCLInitRuleCode;

extern void lclinit_process (void);
extern void lclinit_initMod (void);
extern void lclinit_reset (void);
extern void lclinit_cleanup (void);

