/*
** Copyright (C) University of Virginia, Massachusetts Institue of Technology 1994-2003.
** See ../LICENSE for license information.
**
*/
/*
** help.h
*/

# ifndef HELP_H
# define HELP_H

extern void help_processFlags (int p_argc, char **p_argv) ;
extern void help_showAvailableHelp (void) ;

# else
# error "Multiple include"
# endif





