/*
** Splint - annotation-assisted static program checker
** Copyright (C) 1994-2003 University of Virginia,
**         Massachusetts Institute of Technology
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License as published by the
** Free Software Foundation; either version 2 of the License, or (at your
** option) any later version.
** 
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** General Public License for more details.
** 
** The GNU General Public License is available from http://www.gnu.org/ or
** the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
** MA 02111-1307, USA.
**
** For information on splint: info@splint.org
** To report a bug: splint-bug@splint.org
** For more information: http://www.splint.org
*/
/*
** typeNameNodeList.c
**
** based on list_template.c
**
** where T has T_equal (or change this) and T_unparse
*/

# include "splintMacros.nf"
# include "basic.h"

/*@constant int typeNameNodeListBASESIZE;@*/
# define typeNameNodeListBASESIZE SMALLBASESIZE

typeNameNodeList typeNameNodeList_new (void)
{
  typeNameNodeList s = (typeNameNodeList) dmalloc (sizeof (*s));
  
  s->nelements = 0;
  s->nspace = typeNameNodeListBASESIZE;
  s->elements = (typeNameNode *) 
    dmalloc (sizeof (*s->elements) * typeNameNodeListBASESIZE);

  return (s);
}

static void
typeNameNodeList_grow (typeNameNodeList s)
{
  int i;
  typeNameNode *newelements; 

  s->nspace += typeNameNodeListBASESIZE;

  newelements = (typeNameNode *) dmalloc (sizeof (*newelements)
					  * (s->nelements + s->nspace));

  for (i = 0; i < s->nelements; i++)
    {
      newelements[i] =  s->elements[i]; 
    }

  sfree (s->elements); 
  s->elements = newelements;
}

typeNameNodeList 
typeNameNodeList_add (typeNameNodeList s, typeNameNode el)
{
  if (s->nspace <= 0)
    typeNameNodeList_grow (s);

  s->nspace--;
  s->elements[s->nelements] = el;
  s->nelements++;

  return s;
}

# ifdef DEADCODE
/*@only@*/ cstring
typeNameNodeList_unparse (typeNameNodeList s)
{
  cstring st = cstring_undefined;
  bool first = TRUE;

  typeNameNodeList_elements (s, current)
  {
    if (first)
      {
	st = typeNameNode_unparse (current);
	first = FALSE;
      }
    else
      {
	st = message ("%q, %q", st, typeNameNode_unparse (current));
      }
  } end_typeNameNodeList_elements;
  return (st);
}
# endif /* DEADCODE */

void
typeNameNodeList_free (typeNameNodeList s)
{
  int i;
  for (i = 0; i < s->nelements; i++)
    {
      typeNameNode_free (s->elements[i]); 
    }

  sfree (s->elements);
  sfree (s);
}
