/*
** Copyright (C) University of Virginia, Massachusetts Institue of Technology 1994-2003.
** See ../LICENSE for license information.
**
*/

extern termNode checkSort (/*@returned@*/ termNode p_t);
extern /*@null@*/ termNode computePossibleSorts (/*@returned@*/ /*@null@*/ termNode p_t);

extern void checkLclPredicate (ltoken p_t, lclPredicateNode p_n);
