/*
** Copyright (C) University of Virginia, Massachusetts Institue of Technology 1994-2003.
** See ../LICENSE for license information.
**
*/
typedef struct {
  ltoken tok;
  /*@null@*/ termNode term;
} *arrayQualNode;
